//	Copyright (C) 1999-2003 Core Technologies.
//
//	This file is part of tpasm.
//
//	tpasm is free software; you can redistribute it and/or modify
//	it under the terms of the tpasm LICENSE AGREEMENT.
//
//	tpasm is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	tpasm LICENSE AGREEMENT for more details.
//
//	You should have received a copy of the tpasm LICENSE AGREEMENT
//	along with tpasm; see the file "LICENSE.TXT".


// Generate code for Motorola 6805

#include	"include.h"

static SYMTABLE
	*pseudoOpcodeSymbols,
	*opcodeSymbols;

// enumerated addressing modes

#define	OT_INHERENT				0								// no operands
#define	OT_IMMEDIATE			1								// one byte immediate operand
#define	OT_DIRECT				2								// one byte direct
#define	OT_EXTENDED				3								// two byte absolute
#define	OT_INDIRECT_2			4								// indirect with 2 bytes of offset
#define	OT_INDIRECT_1			5								// indirect with 1 byte of offset
#define	OT_INDIRECT_0			6								// indirect with 0 bytes of offset
#define	OT_RELATIVE				7								// one byte relative offset
#define	OT_BIT_DIRECT			8								// (convenience) one byte of direct, one relative offset
#define	OT_DIRECT_RELATIVE		9								// one byte of direct, one relative offset
#define	OT_BIT_DIRECT_RELATIVE	10								// (convenience) one bit index, one byte of direct, one relative offset

#define	OT_NUM					OT_BIT_DIRECT_RELATIVE+1		// number of addressing modes

// masks for the various addressing modes

#define	M_INHERENT				(1<<OT_INHERENT)
#define	M_IMMEDIATE				(1<<OT_IMMEDIATE)
#define	M_DIRECT				(1<<OT_DIRECT)
#define	M_EXTENDED				(1<<OT_EXTENDED)
#define	M_INDIRECT_2			(1<<OT_INDIRECT_2)
#define	M_INDIRECT_1			(1<<OT_INDIRECT_1)
#define	M_INDIRECT_0			(1<<OT_INDIRECT_0)
#define	M_RELATIVE				(1<<OT_RELATIVE)
#define	M_BIT_DIRECT			(1<<OT_BIT_DIRECT)
#define	M_DIRECT_RELATIVE		(1<<OT_DIRECT_RELATIVE)
#define	M_BIT_DIRECT_RELATIVE	(1<<OT_BIT_DIRECT_RELATIVE)

typedef struct opcode
{
	char
		*name;
	int
		typeMask;
	unsigned char
		baseOpcode[OT_NUM];
} OPCODE;

static PSEUDO_OPCODE
	pseudoOpcodes[]=
	{
		{"db",		HandleDB},
		{"dc.b",	HandleDB},
		{"dw",		HandleBEDW},		// words are big endian
		{"dc.w",	HandleBEDW},
		{"ds",		HandleDS},
		{"ds.b",	HandleDS},
		{"ds.w",	HandleDSW},
		{"incbin",	HandleIncbin},
	};

#define	____		0					// unused opcode

static OPCODE
	Opcodes[]=
	{
//																									inh   imm  dir  ext  ind2 ind1 ind0 rel  bd   dr   bdr
		{"adc",		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xA9,0xB9,0xC9,0xD9,0xE9,0xF9,____,____,____,____}},
		{"add",		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xAB,0xBB,0xCB,0xDB,0xEB,0xFB,____,____,____,____}},
		{"and",		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xA4,0xB4,0xC4,0xD4,0xE4,0xF4,____,____,____,____}},
		{"asl",		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x38,____,____,0x68,0x78,____,____,____,____}},
		{"asla",	M_INHERENT,																		{0x48,____,____,____,____,____,____,____,____,____,____}},
		{"aslx",	M_INHERENT,																		{0x58,____,____,____,____,____,____,____,____,____,____}},
		{"asr",		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x37,____,____,0x67,0x77,____,____,____,____}},
		{"asra",	M_INHERENT,																		{0x47,____,____,____,____,____,____,____,____,____,____}},
		{"asrx",	M_INHERENT,																		{0x57,____,____,____,____,____,____,____,____,____,____}},
		{"bcc",		M_RELATIVE,																		{____,____,____,____,____,____,____,0x24,____,____,____}},
		{"bclr",	M_BIT_DIRECT,																	{____,____,____,____,____,____,____,____,0x11,____,____}},
		{"bclr0",	M_DIRECT,																		{____,____,0x11,____,____,____,____,____,____,____,____}},
		{"bclr1",	M_DIRECT,																		{____,____,0x13,____,____,____,____,____,____,____,____}},
		{"bclr2",	M_DIRECT,																		{____,____,0x15,____,____,____,____,____,____,____,____}},
		{"bclr3",	M_DIRECT,																		{____,____,0x17,____,____,____,____,____,____,____,____}},
		{"bclr4",	M_DIRECT,																		{____,____,0x19,____,____,____,____,____,____,____,____}},
		{"bclr5",	M_DIRECT,																		{____,____,0x1B,____,____,____,____,____,____,____,____}},
		{"bclr6",	M_DIRECT,																		{____,____,0x1D,____,____,____,____,____,____,____,____}},
		{"bclr7",	M_DIRECT,																		{____,____,0x1F,____,____,____,____,____,____,____,____}},
		{"bcs",		M_RELATIVE,																		{____,____,____,____,____,____,____,0x25,____,____,____}},
		{"beq",		M_RELATIVE,																		{____,____,____,____,____,____,____,0x27,____,____,____}},
		{"bhcc",	M_RELATIVE,																		{____,____,____,____,____,____,____,0x28,____,____,____}},
		{"bhcs",	M_RELATIVE,																		{____,____,____,____,____,____,____,0x29,____,____,____}},
		{"bhi",		M_RELATIVE,																		{____,____,____,____,____,____,____,0x22,____,____,____}},
		{"bhs",		M_RELATIVE,																		{____,____,____,____,____,____,____,0x24,____,____,____}},
		{"bih",		M_RELATIVE,																		{____,____,____,____,____,____,____,0x2F,____,____,____}},
		{"bil",		M_RELATIVE,																		{____,____,____,____,____,____,____,0x2E,____,____,____}},
		{"bit",		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xA5,0xB5,0xC5,0xD5,0xE5,0xF5,____,____,____,____}},
		{"blo",		M_RELATIVE,																		{____,____,____,____,____,____,____,0x25,____,____,____}},
		{"bls",		M_RELATIVE,																		{____,____,____,____,____,____,____,0x23,____,____,____}},
		{"bmc",		M_RELATIVE,																		{____,____,____,____,____,____,____,0x2C,____,____,____}},
		{"bmi",		M_RELATIVE,																		{____,____,____,____,____,____,____,0x2B,____,____,____}},
		{"bms",		M_RELATIVE,																		{____,____,____,____,____,____,____,0x2D,____,____,____}},
		{"bne",		M_RELATIVE,																		{____,____,____,____,____,____,____,0x26,____,____,____}},
		{"bpl",		M_RELATIVE,																		{____,____,____,____,____,____,____,0x2A,____,____,____}},
		{"bra",		M_RELATIVE,																		{____,____,____,____,____,____,____,0x20,____,____,____}},
		{"brclr",	M_BIT_DIRECT_RELATIVE,															{____,____,____,____,____,____,____,____,____,____,0x01}},
		{"brclr0",	M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x01,____}},
		{"brclr1",	M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x03,____}},
		{"brclr2",	M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x05,____}},
		{"brclr3",	M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x07,____}},
		{"brclr4",	M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x09,____}},
		{"brclr5",	M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x0B,____}},
		{"brclr6",	M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x0D,____}},
		{"brclr7",	M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x0F,____}},
		{"brset",	M_BIT_DIRECT_RELATIVE,															{____,____,____,____,____,____,____,____,____,____,0x00}},
		{"brset0",	M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x00,____}},
		{"brset1",	M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x02,____}},
		{"brset2",	M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x04,____}},
		{"brset3",	M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x06,____}},
		{"brset4",	M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x08,____}},
		{"brset5",	M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x0A,____}},
		{"brset6",	M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x0C,____}},
		{"brset7",	M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x0E,____}},
		{"brn",		M_RELATIVE,																		{____,____,____,____,____,____,____,0x21,____,____,____}},
		{"bset",	M_BIT_DIRECT,																	{____,____,____,____,____,____,____,____,0x10,____,____}},
		{"bset0",	M_DIRECT,																		{____,____,0x10,____,____,____,____,____,____,____,____}},
		{"bset1",	M_DIRECT,																		{____,____,0x12,____,____,____,____,____,____,____,____}},
		{"bset2",	M_DIRECT,																		{____,____,0x14,____,____,____,____,____,____,____,____}},
		{"bset3",	M_DIRECT,																		{____,____,0x16,____,____,____,____,____,____,____,____}},
		{"bset4",	M_DIRECT,																		{____,____,0x18,____,____,____,____,____,____,____,____}},
		{"bset5",	M_DIRECT,																		{____,____,0x1A,____,____,____,____,____,____,____,____}},
		{"bset6",	M_DIRECT,																		{____,____,0x1C,____,____,____,____,____,____,____,____}},
		{"bset7",	M_DIRECT,																		{____,____,0x1E,____,____,____,____,____,____,____,____}},
		{"bsr",		M_RELATIVE,																		{____,____,____,____,____,____,____,0xAD,____,____,____}},
		{"clc",		M_INHERENT,																		{0x98,____,____,____,____,____,____,____,____,____,____}},
		{"cli",		M_INHERENT,																		{0x9A,____,____,____,____,____,____,____,____,____,____}},
		{"clr",		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x3F,____,____,0x6F,0x7F,____,____,____,____}},
		{"clra",	M_INHERENT,																		{0x4F,____,____,____,____,____,____,____,____,____,____}},
		{"clrx",	M_INHERENT,																		{0x5F,____,____,____,____,____,____,____,____,____,____}},
		{"cmp",		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xA1,0xB1,0xC1,0xD1,0xE1,0xF1,____,____,____,____}},
		{"com",		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x33,____,____,0x63,0x73,____,____,____,____}},
		{"coma",	M_INHERENT,																		{0x43,____,____,____,____,____,____,____,____,____,____}},
		{"comx",	M_INHERENT,																		{0x53,____,____,____,____,____,____,____,____,____,____}},
		{"cpx",		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xA3,0xB3,0xC3,0xD3,0xE3,0xF3,____,____,____,____}},
		{"dec",		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x3A,____,____,0x6A,0x7A,____,____,____,____}},
		{"deca",	M_INHERENT,																		{0x4A,____,____,____,____,____,____,____,____,____,____}},
		{"decx",	M_INHERENT,																		{0x5A,____,____,____,____,____,____,____,____,____,____}},
		{"dex",		M_INHERENT,																		{0x5A,____,____,____,____,____,____,____,____,____,____}},
		{"eor",		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xA8,0xB8,0xC8,0xD8,0xE8,0xF8,____,____,____,____}},
		{"inc",		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x3C,____,____,0x6C,0x7C,____,____,____,____}},
		{"inca",	M_INHERENT,																		{0x4C,____,____,____,____,____,____,____,____,____,____}},
		{"incx",	M_INHERENT,																		{0x5C,____,____,____,____,____,____,____,____,____,____}},
		{"inx",		M_INHERENT,																		{0x5C,____,____,____,____,____,____,____,____,____,____}},
		{"jmp",		M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,						{____,____,0xBC,0xCC,0xDC,0xEC,0xFC,____,____,____,____}},
		{"jsr",		M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,						{____,____,0xBD,0xCD,0xDD,0xED,0xFD,____,____,____,____}},
		{"lda",		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xA6,0xB6,0xC6,0xD6,0xE6,0xF6,____,____,____,____}},
		{"ldx",		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xAE,0xBE,0xCE,0xDE,0xEE,0xFE,____,____,____,____}},
		{"lsl",		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x38,____,____,0x68,0x78,____,____,____,____}},
		{"lsla",	M_INHERENT,																		{0x48,____,____,____,____,____,____,____,____,____,____}},
		{"lslx",	M_INHERENT,																		{0x58,____,____,____,____,____,____,____,____,____,____}},
		{"lsr",		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x34,____,____,0x64,0x74,____,____,____,____}},
		{"lsra",	M_INHERENT,																		{0x44,____,____,____,____,____,____,____,____,____,____}},
		{"lsrx",	M_INHERENT,																		{0x54,____,____,____,____,____,____,____,____,____,____}},
		{"mul",		M_INHERENT,																		{0x42,____,____,____,____,____,____,____,____,____,____}},
		{"neg",		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x30,____,____,0x60,0x70,____,____,____,____}},
		{"nega",	M_INHERENT,																		{0x40,____,____,____,____,____,____,____,____,____,____}},
		{"negx",	M_INHERENT,																		{0x50,____,____,____,____,____,____,____,____,____,____}},
		{"nop",		M_INHERENT,																		{0x9D,____,____,____,____,____,____,____,____,____,____}},
		{"ora",		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xAA,0xBA,0xCA,0xDA,0xEA,0xFA,____,____,____,____}},
		{"rol",		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x39,____,____,0x69,0x79,____,____,____,____}},
		{"rola",	M_INHERENT,																		{0x49,____,____,____,____,____,____,____,____,____,____}},
		{"rolx",	M_INHERENT,																		{0x59,____,____,____,____,____,____,____,____,____,____}},
		{"ror",		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x36,____,____,0x66,0x76,____,____,____,____}},
		{"rora",	M_INHERENT,																		{0x46,____,____,____,____,____,____,____,____,____,____}},
		{"rorx",	M_INHERENT,																		{0x56,____,____,____,____,____,____,____,____,____,____}},
		{"rsp",		M_INHERENT,																		{0x9C,____,____,____,____,____,____,____,____,____,____}},
		{"rti",		M_INHERENT,																		{0x80,____,____,____,____,____,____,____,____,____,____}},
		{"rts",		M_INHERENT,																		{0x81,____,____,____,____,____,____,____,____,____,____}},
		{"sbc",		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xA2,0xB2,0xC2,0xD2,0xE2,0xF2,____,____,____,____}},
		{"sec",		M_INHERENT,																		{0x99,____,____,____,____,____,____,____,____,____,____}},
		{"sei",		M_INHERENT,																		{0x9B,____,____,____,____,____,____,____,____,____,____}},
		{"sta",		M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,						{____,____,0xB7,0xC7,0xD7,0xE7,0xF7,____,____,____,____}},
		{"stop",	M_INHERENT,																		{0x8E,____,____,____,____,____,____,____,____,____,____}},
		{"stx",		M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,						{____,____,0xBF,0xCF,0xDF,0xEF,0xFF,____,____,____,____}},
		{"sub",		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xA0,0xB0,0xC0,0xD0,0xE0,0xF0,____,____,____,____}},
		{"swi",		M_INHERENT,																		{0x83,____,____,____,____,____,____,____,____,____,____}},
		{"tax",		M_INHERENT,																		{0x97,____,____,____,____,____,____,____,____,____,____}},
		{"tst",		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x3D,____,____,0x6D,0x7D,____,____,____,____}},
		{"tsta",	M_INHERENT,																		{0x4D,____,____,____,____,____,____,____,____,____,____}},
		{"tstx",	M_INHERENT,																		{0x5D,____,____,____,____,____,____,____,____,____,____}},
		{"txa",		M_INHERENT,																		{0x9F,____,____,____,____,____,____,____,____,____,____}},
		{"wait",	M_INHERENT,																		{0x8F,____,____,____,____,____,____,____,____,____,____}},
	};

static bool ParseImmediatePreamble(char *theLine,int *lineIndex)
// Expect a pound sign, step over one if found
{
	if(theLine[*lineIndex]=='#')			// does this look like a '#'?
	{
		(*lineIndex)++;						// step over it
		return(true);
	}
	return(false);
}

static bool ParseIndex(char *theLine,int *lineIndex)
// see if next thing on the line is an index register
// return true if so, false otherwise
{
	if(theLine[*lineIndex]=='X'||theLine[*lineIndex]=='x')
	{
		if(!IsLabelChar(theLine[(*lineIndex)+1]))
		{
			(*lineIndex)++;						// step over the register
			return(true);
		}
	}
	return(false);
}

// opcode handling for 6805

enum
{
	POT_INDEX,
	POT_IMMEDIATE,
	POT_VALUE,
};

static bool ParseOperandElement(char *theLine,int *lineIndex,int *type,int *value,bool *unresolved)
// Try to parse an element of the operand and determine its type
// return true if the parsing succeeds
{
	if(ParseImmediatePreamble(theLine,lineIndex))		// an immediate operand?
	{
		if(ParseExpression(theLine,lineIndex,value,unresolved))
		{
			*type=POT_IMMEDIATE;
			return(true);
		}
	}
	else if(ParseIndex(theLine,lineIndex))
	{
		*type=POT_INDEX;
		return(true);
	}
	else if(ParseExpression(theLine,lineIndex,value,unresolved))
	{
		*type=POT_VALUE;
		return(true);
	}
	return(false);
}

static bool HandleImmediate(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// an immediate addressing mode was located
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_IMMEDIATE)
	{
		CheckByteRange(value,true,true);
		if(GenerateByte(theOpcode->baseOpcode[OT_IMMEDIATE],listingRecord))
		{
			fail=!GenerateByte(value,listingRecord);
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleDirect(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// deal with direct mode output only
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_DIRECT)
	{
		CheckUnsignedByteRange(value,true,true);
		if(GenerateByte(theOpcode->baseOpcode[OT_DIRECT],listingRecord))
		{
			fail=!GenerateByte(value,listingRecord);
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleExtended(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// deal with extended mode output only
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_EXTENDED)
	{
		CheckUnsignedWordRange(value,true,true);
		if(GenerateByte(theOpcode->baseOpcode[OT_EXTENDED],listingRecord))
		{
			if(GenerateByte(value>>8,listingRecord))
			{
				fail=!GenerateByte(value&0xFF,listingRecord);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleDirectOrExtended(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// a direct or extended address has been parsed
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_DIRECT)
	{
		if(((value>=0)&&(value<256))||!(theOpcode->typeMask&M_EXTENDED))
		{
			fail=!HandleDirect(theOpcode,value,unresolved,listingRecord);
		}
		else
		{
			fail=!HandleExtended(theOpcode,value,unresolved,listingRecord);
		}
	}
	else
	{
		fail=!HandleExtended(theOpcode,value,unresolved,listingRecord);
	}
	return(!fail);
}

static bool HandleIndirect0(OPCODE *theOpcode,LISTING_RECORD *listingRecord)
// an indirect 0 addressing mode was located
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_INDIRECT_0)
	{
		fail=!GenerateByte(theOpcode->baseOpcode[OT_INDIRECT_0],listingRecord);
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleIndirect1(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// an indirect 1 addressing mode was located
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_INDIRECT_1)
	{
		CheckUnsignedByteRange(value,true,true);
		if(GenerateByte(theOpcode->baseOpcode[OT_INDIRECT_1],listingRecord))
		{
			fail=!GenerateByte(value,listingRecord);
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleIndirect2(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// an indirect 2 addressing mode was located
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_INDIRECT_2)
	{
		CheckUnsignedWordRange(value,true,true);
		if(GenerateByte(theOpcode->baseOpcode[OT_INDIRECT_2],listingRecord))
		{
			if(GenerateByte(value>>8,listingRecord))
			{
				fail=!GenerateByte(value&0xFF,listingRecord);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleIndirect1or2(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// an indirect 1 or 2 addressing mode was located
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_INDIRECT_1)
	{
		if(((value>=0)&&(value<256))||!(theOpcode->typeMask&M_INDIRECT_2))
		{
			fail=!HandleIndirect1(theOpcode,value,unresolved,listingRecord);
		}
		else
		{
			fail=!HandleIndirect2(theOpcode,value,unresolved,listingRecord);
		}
	}
	else
	{
		fail=!HandleIndirect2(theOpcode,value,unresolved,listingRecord);
	}
	return(!fail);
}

static bool HandleBitDirect(OPCODE *theOpcode,int value1,bool unresolved1,int value2,bool unresolved2,LISTING_RECORD *listingRecord)
// a bit direct operand has been parsed
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_BIT_DIRECT)
	{
		Check8BitIndexRange(value1,true,true);
		CheckUnsignedByteRange(value2,true,true);
		value1&=0x07;
		if(GenerateByte(theOpcode->baseOpcode[OT_BIT_DIRECT]|(value1<<1),listingRecord))
		{
			fail=!GenerateByte(value2,listingRecord);
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleDirectRelative(OPCODE *theOpcode,int value1,bool unresolved1,int value2,bool unresolved2,LISTING_RECORD *listingRecord)
// a direct relative operand has been parsed
{
	int
		offset;
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_DIRECT_RELATIVE)
	{
		if(GenerateByte(theOpcode->baseOpcode[OT_DIRECT_RELATIVE],listingRecord))
		{
			CheckUnsignedByteRange(value1,true,true);
			if(GenerateByte(value1,listingRecord))
			{
				offset=0;
				if(!unresolved2&&currentSegment)
				{
					offset=value2-(currentSegment->currentPC+currentSegment->codeGenOffset)-1;
					Check8RelativeRange(offset,true,true);
				}
				fail=!GenerateByte(offset,listingRecord);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleBitDirectRelative(OPCODE *theOpcode,int value1,bool unresolved1,int value2,bool unresolved2,int value3,bool unresolved3,LISTING_RECORD *listingRecord)
// a bit direct relative operand has been parsed
{
	int
		offset;
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_BIT_DIRECT_RELATIVE)
	{
		Check8BitIndexRange(value1,true,true);
		value1&=0x07;
		if(GenerateByte(theOpcode->baseOpcode[OT_BIT_DIRECT_RELATIVE]|(value1<<1),listingRecord))
		{
			CheckUnsignedByteRange(value2,true,true);
			if(GenerateByte(value2,listingRecord))
			{
				offset=0;
				if(!unresolved3&&currentSegment)
				{
					offset=value3-(currentSegment->currentPC+currentSegment->codeGenOffset)-1;
					Check8RelativeRange(offset,true,true);
				}
				fail=!GenerateByte(offset,listingRecord);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleRelative(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// a relative operand has been parsed
{
	bool
		fail;
	int
		offset;

	fail=false;
	if(GenerateByte(theOpcode->baseOpcode[OT_RELATIVE],listingRecord))
	{
		offset=0;
		if(!unresolved&&currentSegment)
		{
			offset=value-(currentSegment->currentPC+currentSegment->codeGenOffset)-1;
			Check8RelativeRange(offset,true,true);
		}
		fail=!GenerateByte(offset,listingRecord);
	}
	else
	{
		fail=true;
	}
	return(!fail);
}

static bool HandleSingleAddress(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// a lone address has been parsed
// This means it is either a relative address, direct, or extended
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_RELATIVE)
	{
		fail=!HandleRelative(theOpcode,value,unresolved,listingRecord);
	}
	else if(theOpcode->typeMask&(M_DIRECT|M_EXTENDED))
	{
		fail=!HandleDirectOrExtended(theOpcode,value,unresolved,listingRecord);
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleDoubleAddress(OPCODE *theOpcode,int value1,bool unresolved1,int value2,bool unresolved2,LISTING_RECORD *listingRecord)
// a dual address has been parsed
// This means it is either BIT_DIRECT, or DIRECT_RELATIVE
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_BIT_DIRECT)
	{
		fail=!HandleBitDirect(theOpcode,value1,unresolved1,value2,unresolved2,listingRecord);
	}
	else if(theOpcode->typeMask&M_DIRECT_RELATIVE)
	{
		fail=!HandleDirectRelative(theOpcode,value1,unresolved1,value2,unresolved2,listingRecord);
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleTripleAddress(OPCODE *theOpcode,int value1,bool unresolved1,int value2,bool unresolved2,int value3,bool unresolved3,LISTING_RECORD *listingRecord)
// a triple address has been parsed
// This means it is BIT_DIRECT_RELATIVE
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_BIT_DIRECT_RELATIVE)
	{
		fail=!HandleBitDirectRelative(theOpcode,value1,unresolved1,value2,unresolved2,value3,unresolved3,listingRecord);
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleSecondValue(OPCODE *theOpcode,char *theLine,int *lineIndex,int value1,bool unresolved1,int value2,bool unresolved2,LISTING_RECORD *listingRecord)
// there are two value, and something still remains
// the only 3 valued operand type is BIT_DIRECT_RELATIVE
// so parse and apply
{
	bool
		fail;
	int
		elementType;
	int
		value3;
	bool
		unresolved3;

	fail=false;
	if(ParseCommaSeparator(theLine,lineIndex))		// make sure the next thing separates operands
	{
		if(ParseOperandElement(theLine,lineIndex,&elementType,&value3,&unresolved3))
		{
			if(elementType==POT_VALUE)
			{
				if(ParseComment(theLine,lineIndex))
				{
					return(HandleTripleAddress(theOpcode,value1,unresolved1,value2,unresolved2,value3,unresolved3,listingRecord));
				}
				else
				{
					ReportBadOperands();
				}
			}
			else
			{
				ReportBadOperands();
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleFirstValue(OPCODE *theOpcode,char *theLine,int *lineIndex,int value1,bool unresolved1,LISTING_RECORD *listingRecord)
// a non-immediate value was parsed, so see what else we can find
{
	bool
		fail;
	int
		elementType;
	int
		value2;
	bool
		unresolved2;

	fail=false;
	if(ParseCommaSeparator(theLine,lineIndex))		// make sure the next thing separates operands
	{
		if(ParseOperandElement(theLine,lineIndex,&elementType,&value2,&unresolved2))
		{
			if(elementType==POT_INDEX)
			{
				if(ParseComment(theLine,lineIndex))
				{
					return(HandleIndirect1or2(theOpcode,value1,unresolved1,listingRecord));
				}
				else
				{
					ReportBadOperands();
				}
			}
			else if(elementType==POT_VALUE)
			{
				if(ParseComment(theLine,lineIndex))
				{
					return(HandleDoubleAddress(theOpcode,value1,unresolved1,value2,unresolved2,listingRecord));
				}
				else
				{
					return(HandleSecondValue(theOpcode,theLine,lineIndex,value1,unresolved1,value2,unresolved2,listingRecord));
				}
			}
			else
			{
				ReportBadOperands();
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleOpcode(void *theOpcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// look at the type of opcode available, parse operands as allowed
// return false only if there was a 'hard' error
{
	OPCODE
		*actualOpcode;
	bool
		fail;
	int
		elementType;
	int
		value;
	bool
		unresolved;

	fail=false;
	actualOpcode=(OPCODE *)theOpcode;
	if(!ParseComment(theLine,lineIndex))
	{
		if(ParseCommaSeparator(theLine,lineIndex))	// see if initial separator (allowed if next thing is an index register)
		{
			if(ParseOperandElement(theLine,lineIndex,&elementType,&value,&unresolved))
			{
				if(elementType==POT_INDEX)
				{
					if(ParseComment(theLine,lineIndex))
					{
						return(HandleIndirect0(actualOpcode,listingRecord));
					}
					else
					{
						ReportBadOperands();
					}
				}
				else
				{
					ReportBadOperands();
				}
			}
			else
			{
				ReportBadOperands();
			}
		}
		else	// first character of operand was not a ','
		{
			if(ParseOperandElement(theLine,lineIndex,&elementType,&value,&unresolved))
			{
				if(elementType==POT_INDEX)
				{
					if(ParseComment(theLine,lineIndex))
					{
						return(HandleIndirect0(actualOpcode,listingRecord));
					}
					else
					{
						ReportBadOperands();
					}
				}
				else if(elementType==POT_IMMEDIATE)
				{
					if(ParseComment(theLine,lineIndex))
					{
						return(HandleImmediate(actualOpcode,value,unresolved,listingRecord));
					}
					else
					{
						ReportBadOperands();
					}
				}
				else if(elementType==POT_VALUE)
				{
					if(ParseComment(theLine,lineIndex))
					{
						return(HandleSingleAddress(actualOpcode,value,unresolved,listingRecord));
					}
					else
					{
						return(HandleFirstValue(actualOpcode,theLine,lineIndex,value,unresolved,listingRecord));
					}
				}
			}
			else
			{
				ReportBadOperands();
			}
		}
	}
	else
	{
		if(actualOpcode->typeMask&M_INHERENT)
		{
			fail=!GenerateByte(actualOpcode->baseOpcode[OT_INHERENT],listingRecord);
		}
		else
		{
			ReportBadOperands();
		}
	}
	return(!fail);
}

static void *MatchOpcode(char *theOpcode)
// match opcodes for this processor, return NULL if none matched
{
	return(STFindDataForName(opcodeSymbols,theOpcode));
}

static bool HandlePseudoOpcode(PSEUDO_OPCODE *theOpcode,char *theLine,int *lineIndex,bool haveLabel,bool isLocal,char *lineLabel,LISTING_RECORD *listingRecord)
// look at the type of opcode available, parse operands as allowed
// return false only if there was a 'hard' error
{
	return(theOpcode->theFunction(theOpcode,theLine,lineIndex,haveLabel,isLocal,lineLabel,listingRecord));
}

static PSEUDO_OPCODE *MatchPseudoOpcode(char *theOpcode)
// match pseudo opcodes for this processor, return NULL if none matched
{
	return((PSEUDO_OPCODE *)STFindDataForName(pseudoOpcodeSymbols,theOpcode));
}

static bool SelectProcessor(PROCESSOR *theProcessor)
// A processor in this family is being selected to assemble with
{
	return(true);
}

static void DeselectProcessor(PROCESSOR *theProcessor)
// A processor in this family is being deselected
{
}

static void UnInitFamily()
// undo what InitFamily did
{
	STDisposeSymbolTable(opcodeSymbols);
	STDisposeSymbolTable(pseudoOpcodeSymbols);
}

static bool InitFamily()
// initialize symbol table
{
	unsigned int
		i;
	bool
		fail;

	fail=false;
	if((pseudoOpcodeSymbols=STNewSymbolTable()))
	{
		for(i=0;!fail&&(i<sizeof(pseudoOpcodes)/sizeof(PSEUDO_OPCODE));i++)
		{
			if(!STAddEntryAtEnd(pseudoOpcodeSymbols,pseudoOpcodes[i].name,&pseudoOpcodes[i]))
			{
				fail=true;
			}
		}
		if(!fail)
		{
			if((opcodeSymbols=STNewSymbolTable()))
			{
				for(i=0;!fail&&(i<sizeof(Opcodes)/sizeof(OPCODE));i++)
				{
					if(!STAddEntryAtEnd(opcodeSymbols,Opcodes[i].name,&Opcodes[i]))
					{
						fail=true;
					}
				}
				if(!fail)
				{
					return(true);
				}
				STDisposeSymbolTable(opcodeSymbols);
			}
		}
		STDisposeSymbolTable(pseudoOpcodeSymbols);
	}
	return(false);
}

// processors handled here (the constuctors for these variables link them to the global
// list of processors that the assembler knows how to handle)

static PROCESSOR_FAMILY
	processorFamily("Motorola 6805",InitFamily,UnInitFamily,SelectProcessor,DeselectProcessor,MatchPseudoOpcode,HandlePseudoOpcode,MatchOpcode,HandleOpcode);

static PROCESSOR
	processors[]=
	{
		PROCESSOR(&processorFamily,"6805",NULL),
		PROCESSOR(&processorFamily,"68705",NULL),
	};
